/*********************************************************************/
/*             Main Control Program Used to Semantic Checking        */ 
/*                   PACT Group, Peking University                   */ 
/*                                                                   */
/*   rewritten by Zheng Gengbin  1996.5.15                           */
/*********************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(STDC_HEADERS) && (STDC_HEADERS==1)
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#ifndef NeXT
#  include <malloc.h>
#endif
 
#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"

#include "defnodes.h"
#include "dep.h"
#include "dep_struct.h"
#include "tags.h"
#include "decl.h"

#include "global.h"
#include "check.h"

int	debug;
int     debug1=0;
int     intrinc_flag=0;
int	cc_flag = 0;                /* if partly compile , set 1 */

int     g_state;
int     scope_status = NON_PURE;
int 	checking_part;
int     check_seq;

int     pre_processing;

SYMB_LIST   rename_list;

PTR_MOD     root;

PTR_TYPE retrieve_global_type();

char    *cur_fname;

static void
usage()
{
	printf("Usage: hpfsc [-vcdoi] HPF_file.f \n");
        printf("       -v    : version          \n");
        printf("       -c    : compile the source files, but do not link. \n");
        printf("       -d    : debugging parser \n");
        printf("       -o    : output filename. \n");
        printf("       -i    : standard library compilation \n");
	exit(1);
}


int main(argc, argv, envp)
    int	  argc;
    char *argv[];
    char *envp[];
{
	FILE	 *fd;
	int i;

/* 	argv++; argc--;   */

	/* Check if there is required .dep FILE name */

	if((argc==1)||(argc>4))
        {
		usage();
        	exit(0);
        }
	else
	{
        argv++;
        while ((argc > 1) && (*argv)[0] == '-')
           {
            if(strcmp(argv[0],"-v")==0) {
		printf("hpfsc version : %s \n",VERSION);
		exit (0);
	    }
            else
            if(strcmp(argv[0],"-d")==0)
                   debug1=1;
            else
	    if (strcmp(argv[0],"-i")==0)
		   intrinc_flag =1;
	    else
            if (strcmp(argv[0],"-c")==0)
		   cc_flag=1;
            else
			usage();
            argc--;
            argv++;
            }		/* while */
        }

	/* Open the specified .dep file */

	cur_fname=*argv;
	if ((fd=fopen(*argv, "r")) == NULL) {
		fprintf(stderr,"Cannot open file \"%s\"\n", *argv);
		exit(1);
	}

	if ((fi = (PTR_FILE) calloc(1, sizeof(struct file_obj))) == NULL) {
		fprintf(stderr, "main -- No more space\n");
		exit(1);
	}

	/* The following program is try to restore  the AST & ST structure
	   from the specified .dep */

	fi->fid = fd;

	if (readNodes(fi) >= 0) {
		fi->global_bfnd->control_parent = NULL;
                semantic_check(); 
		writeNodes(fi, *argv);
			/* create mod_files */
		if (cc_flag)
			write_modules();
	}
	return 0;
}

int semantic_check()
{
     PTR_BLOB program_unit_blob;
     PTR_BFND next_bfnd;

		/* reget global type , no need when merge hpf2dep */
	reget_type();
		/* load intrinsic function lib */
	init_intrinsic_function();	/* add by zheng.gb */
		/* build hash table */
	build_hash(fi->head_symb,hash_table);

		/* preprocess */
	if (!intrinc_flag) preprocess(); 
		/* do initialization */
        initialization();    
		
        cur_bfnd=fi->global_bfnd;  
        program_unit_blob=cur_blob=fi->global_bfnd->entry.Template.bl_ptr1;  
	
        while (program_unit_blob) {

             check_one_program_unit(program_unit_blob);   /*check one program unit*/             

             program_unit_blob=program_unit_blob->next; /*next program unit*/
	}
        
        /*this procedure is used to deal with the check which should be done after scan all the AST*/
        end_processing();       
}

int check_contains(contains_blob)
PTR_BLOB  contains_blob;
{	

        contains_blob = contains_blob->next; /*point to first procedure in conatins*/

        while (contains_blob ) {

             check_one_contain_procedure(contains_blob);   /*check one procedure*/             
             contains_blob=contains_blob->next; /*next program unit*/
	}
}

int reget_type()
{
/*    recover global type  del after join with hpfc */
   global_int      = retrieve_global_type(T_INT);
   global_float    = retrieve_global_type(T_FLOAT);
   global_double   = retrieve_global_type(T_DOUBLE);
   global_char     = retrieve_global_type(T_CHAR);
   global_string   = retrieve_global_type(T_STRING);
   global_bool     = retrieve_global_type(T_BOOL);
   global_complex  = retrieve_global_type(T_COMPLEX);
   global_dcomplex = retrieve_global_type(T_DCOMPLEX);
   global_default  = retrieve_global_type(DEFAULT);
}


int initialization()
{
	PTR_SYMB sym;
	PTR_LABEL label;

if (debug1) printf("***Initialization ! ***\n");
	init_scope_table();
	top_implicit_level=-1;
					/* scope_table start from 1 */
	add_scope_level(fi->global_bfnd,0);

	init_errMsg();

	init_sub();/*initiate function or saubroutine name*/
 	isrecursive_modu(); /*added by Yanmi, module is recursive?*/

	globeErr_g=0;
	default_access=0;
	in_module = 0; /* added by Feng Xiaobing */
	in_block_data = 0; /* added by Feng Xiaobing */
	empty_access_stmt = 0; /* added by Feng Xiaobing */
	in_type_define = 0; /* added by Feng Xiaobing */
	in_interface_body = 0;
	common_table = CNTNULL; /* added by Feng Xiaobing */
	rename_list = (SYMB_LIST)NULL;   
	root = (PTR_MOD)NULL;

			/* reset decl */
	for (sym = fi->head_symb->thread ; sym ; sym=sym->thread) 
	{
		sym->decl = SOFT;
	}
			/* reset label isused */
	for (label = fi->head_lab; label; label=label->next )
	{
		label->labused=0;
	}
}

int check_one_program_unit(unit_blob)           /* program header */
PTR_BLOB unit_blob;
{
	PTR_BLOB cur_blob,contains_blob;

	checking_part=1; /*if checking specification part,checking_part = 1;
			  if checking execution part,checking_part = 2;
			  if checking contain part,checking_part =3;*/

	cur_bfnd=unit_blob->ref;
	check_header(cur_bfnd);
	check_seq = 0xFF ;
        check_specification_stmt(cur_bfnd);  

	cur_blob = cur_bfnd->entry.Template.bl_ptr1;
	contains_blob = cur_bfnd->entry.Template.bl_ptr2;
	
	cur_bfnd = cur_blob->ref;

        while (cur_blob->next && is_specification_stmt(cur_bfnd)) 
	{
            	check_specification_stmt(cur_bfnd);  
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

	if (debug1) printf("*********************%d**********************\n",checking_part);
	if (checking_part == 1) {    /* at intersection of specificatin part*/
		check_at_end_of_spec();			/*and executin  part*/
                checking_part = 2;
        }

	while (cur_blob->next) {
		if (!execution_part()) break;
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

        if (contains_blob) {
		checking_part=3;
    		check_contains(contains_blob);
         }

	cur_bfnd = cur_blob->ref;
	check_end();
}

int check_one_contain_procedure(unit_blob)
PTR_BLOB unit_blob;                        /* procedure header */
{
	PTR_BLOB  cur_blob,contains_blob;
	int       part;

	part=1; /*if checking specification part,checking_part = 1;
			  if checking execution part,checking_part = 2;
			  if checking contain part,checking_part =3;*/

    
	cur_bfnd=unit_blob->ref;
	check_header(cur_bfnd);
	check_seq = 0xFF ;
	cur_bfnd=unit_blob->ref;
        check_specification_stmt(cur_bfnd);  

	cur_blob = cur_bfnd->entry.Template.bl_ptr1;
	contains_blob = cur_bfnd->entry.Template.bl_ptr2;
	
	cur_bfnd = cur_blob->ref;

        while (cur_blob->next && is_specification_stmt(cur_bfnd)) 
	{
            	check_specification_stmt(cur_bfnd);  
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

	if (part == 1){     /* at intersection of specificatin part*/
                check_at_end_of_spec();			/*and executin  part*/
                part = 2;
        }

	while (cur_blob->next) {
		if (!execution_part()) break;
		cur_blob = cur_blob->next;
		cur_bfnd = cur_blob->ref;
	}

        if (contains_blob) {
		part=3;
    		check_contains(contains_blob);
         }

	cur_bfnd = cur_blob->ref;
	check_end();
}

int check_header(cur_bfnd)
PTR_BFND cur_bfnd;
{
	PTR_LLND interface_list;
	SYMB_LIST tmp;
	int i;

/*
	add_scope_level(cur_bfnd,0);
	if (cur_bfnd->entry.Template.symbol && in_interface_body==0 )
	{
		interface_list=get_interface_list(cur_bfnd);
		push_interface(interface_list);
	}
*/
	push_scope(cur_bfnd);

	if (top_implicit_level== -1)
	{
	top_implicit_level++;	
	for (i=0;i<27;i++) 
		if (i>= 'i'-'a' && i<= 'n'-'a')
			implicit_table[i][top_implicit_level]=global_int;
		else 
			implicit_table[i][top_implicit_level]=global_float;
	}
	else {
	   top_implicit_level++;
	   for (i=0;i<27;i++)
		implicit_table[i][top_implicit_level]=
				implicit_table[i][top_implicit_level-1];
	}
	implicit_flag = 0; /* added by Feng Xiaobing */
	save_flag = 0; /* added by Feng Xiaobing */
	empty_save_flag = 0; /* added by Feng Xiaobing */	
	if (debug1) show_cur_scope();

	while (rename_list) {
		tmp=rename_list;
		rename_list=rename_list->next;
		free(tmp);
	}
	rename_list= (SYMB_LIST) NULL;
}

int is_specification_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	check_stmt_function();			/* check if stmt_function */

	switch (cur_bfnd->variant ){
	case  CONTROL_END:
	case  MODULE_STMT:
	case  PROG_HEDR:
	case  FUNC_HEDR:
	case  PROC_HEDR:
	case  BLOCK_DATA:

	case  PARAM_DECL:
	case  STRUCT_DECL:
	case  VAR_DECL :
	case  SAVE_DECL :  
	case  COMM_STAT :
	case  USE_STMT:
	case  DIM_STAT:
	case  DIRSEQUENCE_STMT :
	case  NOSEQUENCE_STMT :
	case  IMPL_DECL :
	case  EQUI_STAT :
	case  INTENT_STMT :
	case  OPTIONAL_STMT :
	case  ACCESS_STMT :
	case  ALLOCATABLE_STMT :
	case  POINTER_STMT :
	case  TARGET_STMT :
	case  DATA_DECL :
	case  PRIVATE_STMT :
	case  SEQUENCE_STMT :
	case  NAMELIST_STAT :	
	/* by Yanmi*/
	case    INTERFACE_STMT:
	case	EXTERN_STAT:
	case	INTRIN_STAT:
	case 	ENTRY_STAT:
        case    FORMAT_STAT:
        case    MODULE_PROC_STMT:
	case    STMTFN_STAT:
	/* by WZL-LY, 03-12 */
        case    DISTRIBUTE_DECL:
        case    ALIGN_DECL:
        case    COMBINED_DIRECTIVE_DECL:
        case    PROCESSORS_STMT:
        case    TEMPLATE_STMT:
        case    INHERIT_DECL:
        case    DYNAMIC_DECL:
			return 1;
	}
	return 0;
}

int check_specification_stmt(cur_bfnd)
PTR_BFND cur_bfnd;
{
	illegal_stmt_label();		/* illegal llabel if there is */

	switch (cur_bfnd->variant ){
	case  CONTROL_END:
			control_end();
			break;
/* added by Yanmi*/
	case   USE_STMT:	
			if ((check_seq & HAS_USE)==0) error1(12);
			use_stmt();
			break;
	case	INTERFACE_STMT:
					/* add by zheng.gb  4.29 1996 */
			check_seq = 0xFF ;
			interface_stmt();
			interface_block();         /* by zheng.gb 1996.5.13*/
			break;
	case	MODULE_PROC_STMT:
			module_proc_stmt();
                       break;
	case	EXTERN_STAT:
			external_stmt();
			break;
	case	INTRIN_STAT:
			intrinsic_stmt();
			break;
	case    PROG_HEDR:
			program_stmt();
			break;
	case  MODULE_STMT:
			module_stmt();
			break;
	case	FUNC_HEDR:
			check_seq = 0xFF ;
			function_stmt();
			break;
	case	PROC_HEDR:
			check_seq = 0xFF ;
			subroutine_stmt();
			break;
	case	ENTRY_STAT:
			entry_stmt();
				/* need not to check the sequence of stmt */
			return;
        case    FORMAT_STAT:
				/* need not to check the sequence of stmt */
			return;
	case    STMTFN_STAT:
			stmt_function_stmt();
			break;
/* the following statements added by Feng Xiaobing */
	case   	VAR_DECL :
			check_var_decl_stmt();
			break;
	case 	DIM_STAT :
			check_dim_stmt();
			break;
	case 	COMM_STAT :
			check_comm_stmt();
			break;
	case 	SAVE_DECL : 
			check_save_decl_stmt();
			break;
	case 	SEQUENCE_STMT :
			check_sequence_stmt();
			break;
	case 	DIRSEQUENCE_STMT :
			check_dirsequence_stmt();
			break;
	case	NOSEQUENCE_STMT :
			check_nosequence_stmt();
			break;
	case	PARAM_DECL :
			check_param_decl();
			break;
	case	IMPL_DECL : 
			if ((check_seq & HAS_IMPLICIT)==0) error1(13);
			check_impl_decl();
			if (implicit_table[26][top_implicit_level]==NULL &&
				(check_seq & HAS_IMPLICIT_NONE)==0 )
				error1(13);
			break;
	case	EQUI_STAT :
			check_equi_stmt();
			break;
	case	INTENT_STMT :	
			check_intent_stmt();
			break;
	case	OPTIONAL_STMT :	
			check_optional_stmt();
			break;
	case	ACCESS_STMT :	
			check_access_stmt();
			break;
	case	ALLOCATABLE_STMT :
			check_allocatable_stmt();
			break;
	case	POINTER_STMT :	
			check_pointer_stmt();
			break;
	case	TARGET_STMT :
			check_target_stmt();
			break;
	case 	DATA_DECL :
			check_data_decl();
			break;
	case 	STRUCT_DECL :
			check_struct_decl();
			strcut_decl_block();           /* zheng.gb */
			break;
	case	PRIVATE_STMT :
			check_private_stmt();
			break;
	case	NAMELIST_STAT :	
			check_namelist_stat();
			break;
	case	BLOCK_DATA : 
			check_block_data();
			break;
	/* by WZL-LY, 03-12 */
        case    DISTRIBUTE_DECL:
                        distribute_decl_smtc(cur_bfnd);
                        break;
        case    ALIGN_DECL:
                        align_decl_smtc(cur_bfnd);
                        break;
        case    COMBINED_DIRECTIVE_DECL:
                        combined_directive_decl_smtc(cur_bfnd);
                        break;
        case    INHERIT_DECL:
                        inherit_decl_smtc(cur_bfnd);
                        break;
        case    DYNAMIC_DECL:
                        dynamic_decl_smtc(cur_bfnd);
                        break;

	}

	if (is_main_header(cur_bfnd)) return;

	/* USE 
	   IMPLICIT NONE
	   PARAMETER         IMPLICIT
        */

	if ((check_seq & HAS_USE)  && cur_bfnd->variant != USE_STMT)
		check_seq &= ~HAS_USE;
	if ((check_seq & HAS_IMPLICIT_NONE) && (check_seq & HAS_USE) ==0 ) 
	{
		if (cur_bfnd->variant!=IMPL_DECL)
			check_seq &= ~HAS_IMPLICIT_NONE;
		else 
			if (implicit_table[26][top_implicit_level]!=NULL)
				check_seq &= ~HAS_IMPLICIT_NONE;
	}
	if (check_seq & HAS_IMPLICIT)
		if ((check_seq & 3 )== 0)
			if (cur_bfnd->variant!=IMPL_DECL &&
				cur_bfnd->variant!=PARAM_DECL )
			  check_seq &= ~HAS_IMPLICIT;

}


int execution_part()
{
        if (executable_construct()) return 1;
        if (execution_part_construct()) return 1;
	if (cur_bfnd->variant == CONTAINS_STMT ||
		cur_bfnd->variant == CONTROL_END ) return 0;
        error2(10,cur_bfnd->variant);
        cur_bfnd=cur_bfnd->thread;
	return 0;
}

int executable_construct()
{
        if ( action_stmt() )  return 1;

        switch (cur_bfnd->variant)
        {
        case LOGIF_NODE:
        	stmt_label();           /* check statement label */
                if_construct();
                break;
	case SWITCH_NODE:
        	stmt_label();           /* check statement label */
		case_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case FOR_NODE:
        	stmt_label();           /* check statement label */
		do_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case DO_WHILE_NODE:
        	stmt_label();           /* check statement label */
		do_while_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case FORALL_CONSTRUCT_STMT:
        	stmt_label();           /* check statement label */
		forall_construct_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
	case WHERE_BLOCK_STMT:
        	stmt_label();           /* check statement label */
		where_block_stmt();
        	cur_bfnd=cur_bfnd->thread;
		break;
        default:
                return 0;
        }

        return 1;
}

int action_stmt()
{
	PTR_SYMB sub_name;

        switch (cur_bfnd->variant)
        {
        case ASSIGN_STAT:
        		stmt_label();           /* check statement label */
                        check_assign_stmt();
                        break;
        case  POINTER_ASSIGN_STAT:
        		stmt_label();           /* check statement label */
                        check_pointer_assign_stmt();
                        break;
	case ALLOCATE_STMT:
        		stmt_label();           /* check statement label */
			allocate_stmt();
			break;
	case DEALLOCATE_STMT:
        		stmt_label();           /* check statement label */
			deallocate_stmt();
			break;
	case NULLIFY_STMT:
        		stmt_label();           /* check statement label */
			nullify_stmt();
			break;
	case WHERE_NODE:
        		stmt_label();           /* check statement label */
			where_stmt();
			break;
        case LOGIF_STMT:
        		stmt_label();           /* check statement label */
                	if_stmt();
                	break;
        case FORALL_STMT:
        		stmt_label();           /* check statement label */
                	forall_stmt();
                	break;
        case GOTO_NODE:
        		stmt_label();           /* check statement label */
                        goto_stmt();
                        break;
	case COMGOTO_NODE:
        		stmt_label();           /* check statement label */
			comput_goto_stmt();
			break;
	case ASSLAB_STAT:
        		stmt_label();           /* check statement label */
			assign_to_stmt();
			break;
	case ASSGOTO_NODE:
        		stmt_label();           /* check statement label */
			assign_goto_stmt();
			break;
	case ARITHIF_NODE:
        		stmt_label();           /* check statement label */
			arithmetic_if_stmt();
	case CYCLE_STMT:
	case EXIT_STMT:
        		stmt_label();           /* check statement label */
			exit_stmt();
			break;
	case STOP_STAT:
        		stmt_label();           /* check statement label */
			stop_stmt();
			break;
	case PAUSE_NODE:
        		stmt_label();           /* check statement label */
			pause_stmt();
			break;
	case CONTINUE_NODE:
        		stmt_label();           /* check statement label */
			break;
	case RETURN_STAT:
        		stmt_label();           /* check statement label */
                        return_stmt();
                        break;
        case PROC_STAT:
        		stmt_label();           /* check statement label */
			sub_name = cur_bfnd->entry.call.call_exp->
						entry.Template.symbol;
			if (is_intrinsic_call(sub_name)!=0)
				check_intrinsic_func(cur_bfnd->entry.call.call_exp,PROCEDURE_NAME);
			else
                        	call_or_func(cur_bfnd,sub_name, cur_bfnd->entry.call.call_exp->entry.Template.ll_ptr1,PROCEDURE_NAME,1,TYNULL);
                        break;
	case  OPEN_STAT:
			open_close();
			break;
	case  READ_STAT:	
			read_stmt();
			break;
	case WRITE_STAT:
			write_stmt();
			break;
	case PRINT_STAT:	
			print_stmt();
			break;
	case CLOSE_STAT:	
			open_close();
			break;
	case INQUIRE_STAT:
			inquire_stmt();
			break;
	case BACKSPACE_STAT:
	case REWIND_STAT:
	case ENDFILE_STAT:
			backspace_endfile_rewind_stmt();
			break;		

	/* by WZL-LY, 03-12 */
        case DISTRIBUTE_STMT:
	case REDISTRIBUTE_STMT:
	case CBREDISTRIBUTE_STMT:
        		stmt_label();           /* check statement label */
                        redistribute_stmt_smtc(cur_bfnd);
                        break;
        case ALIGN_STMT:
	case REALIGN_STMT:
	case CBREALIGN_STMT:
        		stmt_label();           /* check statement label */
                        realign_stmt_smtc(cur_bfnd);
                        break;

        default:
                        return 0;
        }

        cur_bfnd=cur_bfnd->thread;

        return 1;
}

int execution_part_construct()          /* execution_part_construct */
{
        if ( executable_construct() ) return 1;

        switch (cur_bfnd->variant)
        {
          case ENTRY_STAT:
		illegal_stmt_label();		/* illegal llabel if there is */
                break;
          case FORMAT_STAT:
        	stmt_label();           /* check statement label */
                break;
	  case DATA_DECL:
		illegal_stmt_label();
		check_data_decl();
		break;
          default:
                return 0;
        }
        cur_bfnd=cur_bfnd->thread;
        return 1;
}

int block()
{
        return execution_part_construct();
}

int check_end()
{
	check_control_end();

	check_scope_labels();

	top_implicit_level--;
	delete_beyond_scope_level(0);

	if (debug1) show_cur_scope();
}

int end_processing()
{
	is_ckecked_bfnd();
	isrecursive_sub();
	check_interface();
	is_dup_globle();
	
}


PTR_TYPE retrieve_global_type(var)
int var;
{
	PTR_TYPE  tp,kind;

	for (tp=fi->head_type; tp; tp=tp->thread)
		if (tp->variant == var)
			return tp;
}


show_cur_scope()
{
	int i;
/*	printf("Scope level:%d,line:%d\n",top_scope_level,cur_bfnd->g_line);*/
        for (i=1; i<= top_scope_level ; i++)
	{
		printf("%d:",i);
		if (scope_table[i]->entry.Template.symbol)
			printf("%s ",scope_table[i]->entry.Template.symbol->ident);
		else
			printf("-- ");
	}
	printf(" **** %d **** \n",cur_bfnd->g_line);
}

check_at_end_of_spec()
{
	common_equi_process();
	check_dirsequence_attribute();
	check_common_unity();   
}


/*
 *    PREPROCESS() ---  preprocess the AST to decide dummy type .
 */
int preprocess()
{
	PTR_BLOB blob;
	PTR_BFND bfnd;

printf("Preprocessing ...\n");
	pre_processing = 1;
	load_modules();
	blob= fi->global_bfnd->entry.Template.bl_ptr1;
	for ( ; blob ; blob=blob->next) {
		bfnd = blob->ref;
		pre_0(bfnd);
	}
	pre_processing = 0;
printf("Preprocessing done. \n");
}



int pre_0(bfnd)
PTR_BFND bfnd;
{
	PTR_BLOB contains_blob;

	cur_bfnd = bfnd;
	contains_blob = cur_bfnd->entry.Template.bl_ptr2;

	push_scope(cur_bfnd);
/*
	add_scope_level(cur_bfnd,0);
	if (cur_bfnd->entry.Template.symbol && in_interface_body==0 )
	  push_interface(cur_bfnd->entry.Template.symbol->entry.Template.interface_list);
*/

	if (cur_bfnd->variant == FUNC_HEDR || cur_bfnd->variant == PROC_HEDR)
	{
		check_dummy_type(cur_bfnd->entry.Template.symbol);
		check_interface_dummy(cur_bfnd);
	}

	if (contains_blob) {
		for (contains_blob=contains_blob->next;contains_blob;
			contains_blob=contains_blob->next)
		pre_0(contains_blob->ref);
	}

	delete_beyond_scope_level(0);
}


check_dummy_type(sub_name)
PTR_SYMB sub_name;
{
	PTR_SYMB dummy,sym;
	PTR_TYPE type , base_type,default_type;

	for (dummy=sub_name->entry.proc_decl.symb_list; dummy ;
		dummy=dummy->entry.var_decl.next_in)
	{
		type = dummy->type;
		if (type == TYNULL) continue;

        	base_type=get_base_type(type);
        	if (base_type->variant == DEFAULT)
        	{
                	sym=search_symb_ref(base_type->name);
                	if (sym != dummy->type->name &&
                        	sym->variant == TYPE_NAME ) {
                        	default_type = look_up_record(sym);
                        	dummy->type=default_type;
                	}
                }
	}
}


int check_interface_dummy(bfnd)
PTR_BFND bfnd;
{
	PTR_LLND  interface_list,ptr;
	PTR_SYMB  symb,dummy,sub_name,vname;
	PTR_BFND  inter_scope;
	extern PTR_SYMB  look_up_symb_at_scope_in_hash();
	
	interface_list=get_interface_list(cur_bfnd);
	if (interface_list == LLNULL) return;
	sub_name=bfnd->entry.Template.symbol;
	
	for (ptr=interface_list; ptr; ptr=ptr->entry.Template.ll_ptr2)
        {
                symb=ptr->entry.Template.ll_ptr1->entry.Template.symbol;
                inter_scope = symb->scope;
		for (dummy=sub_name->entry.proc_decl.symb_list; dummy ;
                	dummy=dummy->entry.var_decl.next_in)
        	{
			vname=look_up_symb_at_scope_in_hash(dummy->ident,inter_scope);
			if (vname)
				dummy->entry.Template.declared_name = vname;
		}
	}
}


/*
 *     load_modules() -- loads all module used by this file 
 */
int load_modules()
{
	PTR_BLOB blob;
	PTR_BFND bfnd;
	PTR_LLND use_list;

	for (blob= fi->global_bfnd->entry.Template.bl_ptr1; blob;
			blob=blob->next) 
	{
		bfnd=blob->ref;
		if (bfnd->variant != MODULE_STMT) continue;
		use_list=bfnd->entry.Template.symbol->entry.Template.use_list;
		for ( ; use_list; use_list=use_list->entry.Template.ll_ptr2)
		{
			if (search_head_name(use_list->entry.Template.ll_ptr1->entry.Template.symbol->ident)==BFNULL)
			insert_module(use_list->entry.Template.ll_ptr1->entry.Template.symbol->ident);
		}
	}
	if (root) load_mod_files(root);
}


int insert_module(mod)
char *mod;
{
	PTR_MOD new_node,tmp;

	if (present(mod,root)) return;
	new_node = (PTR_MOD)chkalloc(sizeof(struct _binary_mod_tree));
	new_node->lp=NULL;
	new_node->rp=NULL;
	new_node->name=mod;
	if (root==(PTR_MOD)NULL)
		root=new_node;
	else {
		for (tmp=root;tmp->rp;tmp=tmp->rp);
		tmp->rp=new_node;
	}
	expand_by_use(new_node);
}


int present(name,node)
char *name;
PTR_MOD node;
{
	if (node==(PTR_MOD)NULL) return 0;
	if (strcasecmp(node->name,name)==0) return 1;
	if (node->lp)
		if (present(name,node->lp)) return 1;
	if (node->rp)
		if (present(name,node->rp)) return 1;
	return 0;
}


int expand_by_use(node)
PTR_MOD node;
{
}




