/*************************************************************************

	readmod.c -- Separate compilation

written by Zheng Gengbin
*************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifdef STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include <stdio.h>

#   include "typedef.h"
#   include "bif.h"
#   include "ll.h"
#   include "symb.h"
#include "defnodes.h"
#include "dep.h"
#include "mod_struct.h"
#include "tags.h"
#include "decl.h"

#include "exglobal.h"


#define NULL_CHECK(BASE,VALUE) ((VALUE) ? (BASE + (VALUE-1)): 0)

/*
 * Local variables
 */

static struct _locs      floc;   /* used to read in preamble "floc" */
static struct _preamble  head;   /* used to read in preamble "head" */
static struct _ll_nd     ll;     /* used to read in ll nodes */
static struct _sym_nd    sym;    /* used to read in symbol nodes */
static struct _typ_nd    typ;    /* used to read in type nodes */

static PTR_BFND head_bfnd, cur_module;
static PTR_LLND head_llnd, cur_llnd;
static PTR_SYMB head_symb, cur_symb;
static PTR_TYPE head_type, cur_type;
static PTR_DEP  head_dep,  cur_dep;
static PTR_LABEL head_lab, cur_lab;
static PTR_CMNT head_cmnt, cur_cmnt;


static char **strtbl;           /* starting address of string table */
static int tmp[2000];       /* temp working area */
static FILE *fd;                /* local copy of file id for the dep file */
static PTR_FILE lfi;
static int need_swap = 0;       /* set to 1 if we need to swap bytes */

static int     ntype=0;
static int     nsymb=0;
static int     nllnd=0;

/********************************************************
 *                      swap_w                          *
 *                                                      *
 *           Swap bytes of one word (2 bytes)           *
 ********************************************************/
static void
swap_w(p)
        char *p;
{
    char c;

    c = *(p+1);
    *(p+1) = *p;
    *p = c;
}


/*------------------------------------------------------*
		read_string
--------------------------------------------------------*/
static char *read_string()
{
	u_short u;
	int n;
	char *s;

	if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0)
            return NULL;
        if (need_swap)
            swap_w((char *)&u);
        n = (int) u;
        if ((s = malloc(n+1)) == NULL) {
            fprintf(stderr, "read_str_tbl: No more space\n");
            exit(1);
        }
        if ((int)fread(s, sizeof(char), n, fd) < 0) /* now the content */
            return NULL;
        *(s+n) = '\0';
	return s;
}

/*------------------------------------------------------*
 *		   read_str_tbl				*
 *							*
 *	Read in the string table in dep file		*
 *------------------------------------------------------*/
static int
read_str_tbl()
{
    int	     i, n, sz;
    u_short  u;
    char    *s;
    char   **cp;

    /*
     * Fast forward to where the string table starts
     */
    if (fseek(fd, floc.strs, 0) < 0)
	return -1;

    /*
     * The first word is the total number of strings in the dep file
     */

    /* get size of string table */
    if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0) 
	return -1;

    if (need_swap)
	swap_w((char *)&u);
    sz = (int) u;
    if((cp = strtbl = (char **)malloc(sz*sizeof(char *))) == NULL) {
	fprintf(stderr, "read_str_tbl: No more space\n");
	exit(1);
    }

    /*
     * Then followed by strings in the form of
     *		-------------------------
     *		| str length | contents	|
     *		-------------------------
     */
    for (i = 0; i < sz; i++) {
        /* get string length */
	if ((int)fread( (char *) &u, sizeof(u_short), 1, fd) < 0) 

	    return -1;
	if (need_swap)
	    swap_w((char *)&u);
	n = (int) u;
	if ((s = malloc(n+1)) == NULL) {
	    fprintf(stderr, "read_str_tbl: No more space\n");
	    exit(1);
	}
	if ((int)fread(s, sizeof(char), n, fd) < 0) /* now the content */
	    return -1;
	*(s+n) = '\0';
	*cp++ = s;
    }
    return 0;
}

/*--------------------------------------------------------------*
 *                      read_preamble                           *
 *          Read in the preamble part of the dep file           *
 *--------------------------------------------------------------*/
static int
read_preamble()
{
	int i;
	char filemagic;
	char *fname;
	long pos;
	u_short mno[2];

    /* The first 8 bytes is the file magic (see /etc/magic) PHB */
	if ((int)fread( (char *) mno, sizeof(u_short), 1, fd) < 0)
        	return -1;
	if (*mno != MAGIC_NO) {      /* Is this a dep file? */
        need_swap = 1;          /* No... */
        swap_w((char *)mno);            /*  ... Maybe we need to swap bytes */
        	if(*tmp != MAGIC_NO) {   /* Try again */
          	fprintf(stderr, "Are you sure this is a legal mod file? %x\n",*tmp);
          	return -2;
        	}
    	}

	fname=read_string();
printf("Fname:%s\n",fname);

	/* Some more data */
	if ((int)fread( (char *) &floc, sizeof(struct _locs), 1, fd) < 0)
        return -1;

	pos=ftell(fd);

	/* Reconstruct the string table first */
	if (read_str_tbl() < 0)
        return -1;

    /* rewind back to the point after "locs" information (8 is filemagic) */
	if (fseek(fd, pos, 0) < 0)
        return -1;

    /*
     * Read in the second part of preamble.  Here we have numbers of
     * all nodes (bif, low level, etc.) for this dep file
     */
    if ((int)fread( (char *) &head, sizeof(struct _preamble), 1, fd) < 0)
        return -1;
    if (need_swap)
        swab((char *)&head, (char *)&head, sizeof(struct _preamble));

	nsymb=head.num_symbs;
        nllnd=head.num_llnds;
        ntype=head.num_types;
printf("nsym:%d,nllnd:%d,ntype:%d.\n",nsymb,nllnd,ntype);

        head_symb = (PTR_SYMB)(nsymb>0? calloc(nsymb, sizeof(struct symb)): NULL);
	head_type = (PTR_TYPE)(ntype>0? calloc(ntype, sizeof( struct data_type)): NULL);
	head_llnd = (PTR_LLND)(nllnd>0? calloc(nllnd, sizeof( struct llnd)): NULL);
}


PTR_BFND
retrieve_bfnd(no)
int no;
{
	if (no==0) return lfi->global_bfnd;
	else return cur_module;
}


/*
 *   get symbol by the number 
 */
PTR_SYMB
retrieve_sym(no)
int no;
{
	if (no==0) return SMNULL;
	if (no>0)
		return head_symb+(no-1);
	return SMNULL;
}


PTR_TYPE
retrieve_type(no)
int no;
{
	PTR_SYMB type_name;

	switch (no) {
        case T_INT:
		return global_int;
        case T_FLOAT:
		return global_float;
        case T_DOUBLE:
		return global_double;
        case T_COMPLEX:
		return global_complex;
        case T_DCOMPLEX:
		return global_dcomplex;
        case T_BOOL:
		return global_bool;
        case T_STRING:
		return global_string;
        }
	if (no>0)  {
printf("HERENO:%d\n",no);
		no--;
		return head_type+no;
	}
	return TYNULL;
}

/*----------------------------------------------------------------------*
 *									*
 *		    routines to read type table				*
 *									*
 *----------------------------------------------------------------------*/
static int
read_mod_type_nodes()
{
    PTR_TYPE type_ptr;
    int i, uss1, uss2;

    for(i = 0; i < ntype; i++) {
	if ((int)fread( (char *) &typ, sizeof(struct _typ_nd), 1, fd) < 0)
	    return -1;
	if (need_swap)
	    swab((char *)&typ, (char *)&typ, sizeof(struct _typ_nd));

	type_ptr		= head_type + i;
	type_ptr->variant	= (int)typ.variant;
	type_ptr->name		= retrieve_sym(typ.name);
printf("Record name:%d,%s\n",typ.name,type_ptr->name->ident);

	switch (type_ptr->variant) {
	case T_INT	:
	case T_FLOAT	:
	case T_DOUBLE	:
	case T_CHAR	:
	case T_BOOL	:
        case T_COMPLEX  :
        case T_DCOMPLEX  :
	case T_STRING	:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp); 
		/* swab((char *)tmp, (char *)tmp, sizeof(u_short)); */
	    type_ptr->entry.Template.ranges    = NULL_CHECK(head_llnd,tmp[0]);
	    break;
	case DEFAULT	:
	case T_VOID	:	/* NEW ADDED FOR VPC */
        case T_UNKNOWN  :
	case T_ENUM_FIELD:
	    break;

/**  Modify for HPF  2  **/

	case T_ARRAY	:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 3*sizeof(int));
	    type_ptr->entry.ar_decl.num_dimensions = (int)tmp[0];
	    type_ptr->entry.ar_decl.base_type = NULL_CHECK(head_type,tmp[1]);
	    type_ptr->entry.ar_decl.ranges    = NULL_CHECK(head_llnd,tmp[2]);
	    break;

/**  Modify for HPF  1  **/
					/* add by zheng.gb */
	case T_KIND:			
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 2*sizeof(u_short));
	    type_ptr->entry.kind_decl.base_type = NULL_CHECK(head_type,tmp[0]);
	    type_ptr->entry.kind_decl.kindexp = NULL_CHECK(head_llnd,tmp[1]);
	    break;
/**  Modify for HPF  2  **/

	case T_RECORD	:
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 2*sizeof(u_short));
	    type_ptr->entry.re_decl.num_fields = (int)(*tmp);
	    type_ptr->entry.re_decl.first      = retrieve_sym(tmp[1]);
printf("record:%d,%s\n",tmp[1],type_ptr->entry.re_decl.first->ident);
	    break;


	default :
	    break;
	}
    }
    return 0;
}


/*--------------------------------------------------------------*
 *								*
 *		  routines to read symbol table			*
 *								*
 *--------------------------------------------------------------*/
static int
read_symb_nodes()
{
    PTR_SYMB symb_ptr;
    int i;

    for(i = 0; i < nsymb; i++) {
	if ((int)fread( (char *) &sym, sizeof(struct _sym_nd), 1, fd) < 0)
	    return -1;
	if (need_swap)
	    swab((char *)&sym, (char *)&sym, sizeof(struct _sym_nd));

	symb_ptr	    = head_symb + i;
	symb_ptr->variant   = (int) sym.variant;
	symb_ptr->type	    = retrieve_type(sym.type);
printf("retrived type:%d,%s.\n",sym.type,symb_ptr->type->name->ident);
	symb_ptr->attr	    = (int) sym.attr;
	symb_ptr->next_symb = SMNULL; /*retrieve_sym(sym.next);*/
	symb_ptr->scope	    = retrieve_bfnd(i);
	symb_ptr->ident	    = *(strtbl + sym.ident);

	switch (symb_ptr->variant) {
	case DEFAULT	:
	case TYPE_NAME	:
	    break;
	case CONST_NAME	:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    if (need_swap)
		swap_w((char *)tmp);
	    symb_ptr->entry.const_value = LLNULL;/*NULL_CHECK(head_llnd,(*tmp));*/
	    break;
	case ENUM_NAME	:
	case FIELD_NAME	:
	    if ((int)fread( (char *) tmp, sizeof(int), 3, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 5*sizeof(int));
	    symb_ptr->entry.field.tag	= (int)(*tmp);
	    symb_ptr->entry.field.next	= NULL_CHECK(head_symb,tmp[1]);
	    symb_ptr->entry.field.base_name= NULL_CHECK(head_symb,tmp[2]);
/**  Modify for HPF 1	    symb_ptr->entry.field.declared_name = NULL_CHECK(head_symb,tmp[3]);  **/
/**  Modify for HPF 2	    symb_ptr->entry.field.restricted_bit= NULL_CHECK(head_llnd,tmp[4]);  **/
	    break;
	case VARIABLE_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 4, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 4*sizeof(int));
	    symb_ptr->entry.var_decl.local  = (int)(*tmp);
	    symb_ptr->entry.var_decl.next_in= retrieve_sym(tmp[1]);
	    symb_ptr->entry.var_decl.next_out=retrieve_sym(tmp[2]);
/* 		add by zheng.gb */
	    symb_ptr->entry.Template.declared_name=retrieve_sym(tmp[3]);
	    break;
	case PROGRAM_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 4, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 4*sizeof(int));

	    symb_ptr->entry.prog_decl.symb_list = NULL_CHECK(head_symb,(*tmp));
	    symb_ptr->entry.prog_decl.prog_hedr = NULL_CHECK(head_bfnd,tmp[1]);
	    symb_ptr->entry.prog_decl.use_list = NULL_CHECK(head_llnd,tmp[2]);
	    symb_ptr->entry.prog_decl.interface_list = NULL_CHECK(head_llnd,tmp[3]);
	    break;
	case MODULE_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 2, fd) < 0)
                return -1;
	    if (need_swap)
                swab((char *)tmp, (char *)tmp, 2*sizeof(int));
            symb_ptr->entry.prog_decl.use_list = LLNULL;/*NULL_CHECK(head_llnd,tmp[0]);*/
            symb_ptr->entry.prog_decl.interface_list = LLNULL;/*NULL_CHECK(head_llnd,tmp[1]);*/
            break;
	case PROCEDURE_NAME :
        case PROCESS_NAME:
	case FUNCTION_NAME:	 
	    if ((int)fread( (char *) tmp, sizeof(int),11, fd) < 0)
		return -1;
	    if (need_swap)
		swab((char *)tmp, (char *)tmp, 11*sizeof(int));

	    symb_ptr->entry.proc_decl.num_input	 = (int)(*tmp);
	    symb_ptr->entry.proc_decl.num_output = (int)tmp[1];
	    symb_ptr->entry.proc_decl.num_io	 = (int)tmp[2];
	    symb_ptr->entry.proc_decl.in_list =NULL_CHECK(head_symb,tmp[3]);
	    symb_ptr->entry.proc_decl.out_list =NULL_CHECK(head_symb,tmp[4]);
	    symb_ptr->entry.proc_decl.symb_list=NULL_CHECK(head_symb,tmp[5]);
	    symb_ptr->entry.proc_decl.proc_hedr=NULL_CHECK(head_bfnd,tmp[6]);
            symb_ptr->entry.proc_decl.local_size = (int)tmp[7];
	    symb_ptr->entry.proc_decl.use_list=NULL_CHECK(head_llnd,tmp[8]);
	    symb_ptr->entry.proc_decl.interface_list=NULL_CHECK(head_llnd,tmp[9]);
	    symb_ptr->entry.Template.declared_name=NULL_CHECK(head_symb,tmp[10]);
	    break;
	case INTERFACE_NAME:
	    if ((int)fread( (char *) tmp, sizeof(int), 1, fd) < 0)
		return -1;
	    symb_ptr->entry.Template.declared_name=NULL_CHECK(head_symb,tmp[0]);
	    break;
	default:
	    break;
	}
    }
    return 0;
}

int
read_mod_nodes(fi,name)
PTR_FILE fi;
char *name;
{
	PTR_BFND module;
	PTR_BLOB blob,mod_blob;
	PTR_BFND bfnd,end;
	PTR_SYMB sym;
	PTR_TYPE type;
	int      i;
	extern PTR_HASH  hash_table[];

	printf("Loading module file '%s'...\n",name);

	lfi=fi;

	if ((fd = fopen (name, "r")) == NULL) {
                fprintf(stderr, "Could not open %s for read\n", name);
                return (-1);
        }
	
	if (read_preamble() < 0)
        	return -1;
	
	module=makeBfnd(fi,MODULE_STMT);
	module->control_parent=fi->global_bfnd;
	/* link to the BFND */
	for (bfnd=fi->head_bfnd;bfnd->thread; bfnd=bfnd->thread);
	bfnd->thread=module;
	module->thread=BFNULL;

	mod_blob=make_blob(fi,module,BLNULL);
	for (blob=fi->global_bfnd->entry.Template.bl_ptr1;blob->next; blob=blob->next);
	blob->next=mod_blob;
		/* control_end */
	end=makeBfnd(fi,CONTROL_END);
	end->control_parent=module;
	module->thread=end;
	mod_blob=make_blob(fi,end,BLNULL);
	bfnd->entry.Template.bl_ptr1=mod_blob;

	cur_module=bfnd;
	read_symb_nodes();

	module->entry.Template.symbol=head_symb;

	/*  merge */
	fi->num_symbs += nsymb;
	for (sym=fi->head_symb; sym->thread; sym=sym->thread);
	for (i = 0; i < nsymb; i++) {
        	(head_symb + i)->id     = i + 1 + sym->id;
        	(head_symb + i)->thread=(i==nsymb-1)?SMNULL:head_symb + (i + 1);
		/* insert into hash table */
		insert_hash(head_symb+i,hash_table);
        }
	sym->thread=head_symb;
	for (sym=fi->head_symb; sym->next_symb; sym=sym->next_symb);
	sym->next_symb=head_symb;

	/* begin to read type table */
	if (ftell(fd) != floc.type) {
        	fprintf(stderr,"readNodes: wrong location of type nodes\n");
        	if(fseek(fd, floc.type, 0) < 0)
            	return -1;
    	}

    	if (read_mod_type_nodes() < 0)
        	return -1;

	merge_type_at_end(fi,head_type,ntype);
}



int merge_type_at_end(fi,new_head,ntype)
PTR_FILE fi;
PTR_TYPE new_head;
int ntype;
{
	PTR_TYPE type;
	int 	i;

	/* merge type */
	fi->num_types += ntype;
	for (type=fi->head_type; type->thread; type=type->thread);
	for (i = 0; i < ntype; i++) {
        	(new_head + i)->id     = i + 1 +type->id;
        	(new_head + i)->thread=(i==ntype-1)?TYNULL:new_head + (i + 1);
    	}
	type->thread=new_head;
}



